'''
Created on Mar 27, 2017

@author: nicolas
'''

from lemoncheesecake.utils import is_string
from lemoncheesecake.runtime import log_check
from lemoncheesecake.exceptions import AbortTest, ProgrammingError
from lemoncheesecake.matching.matchers.dict_ import HasEntry, wrap_key_matcher
from lemoncheesecake.matching.matchers.composites import is_

__all__ = (
    "log_match_result",
    "check_that", "check_that_entry",
    "require_that", "require_that_entry",
    "assert_that", "assert_that_entry",
    "this_dict"
)


class _HasEntry(HasEntry):
    def description(self):
        ret = 'entry %s' % self.key_matcher.description()
        if self.value_matcher:
            ret += " " + self.value_matcher.description()
        return ret


class DictContext:
    def __init__(self, actual, base_key):
        self.actual = actual
        self.base_key = base_key


class ThisDict:
    _contexts = []

    def __init__(self, actual):
        self.context = DictContext(actual, [])

    def using_base_key(self, base_key):
        self.context.base_key = base_key if type(base_key) in (list, tuple) else [base_key]
        return self

    def __enter__(self):
        self._contexts.append(self.context)
        return self.context.actual

    def __exit__(self, type_, value, traceback):
        self._contexts.pop()

    @staticmethod
    def get_current_context():
        try:
            return ThisDict._contexts[-1]
        except IndexError:
            return None


def this_dict(actual):
    """
    Set actual to be used as implict dict in *_that_entry functions when used
    within a 'with' statement.
    """
    return ThisDict(actual)


def _entry_operation(operation):
    def wrapper(key_matcher, value_matcher=None, in_=None, quiet=False):
        actual = in_
        base_key = []
        if not actual:
            context = ThisDict.get_current_context()
            if not context:
                raise ProgrammingError("Actual dict must be set either using in_ argument or using 'with this_dict(...)' statement")
            actual = context.actual
            base_key = context.base_key
        matcher = _HasEntry(
            wrap_key_matcher(key_matcher, base_key=base_key),
            value_matcher if value_matcher != None else is_(value_matcher)
        )
        return operation("", actual, matcher, quiet=quiet)
    wrapper.__doc__ = "Helper function for %s, takes the actual dict using in_ parameter or using 'with this_dict(...)' statement" % \
        operation.__name__
    return wrapper


def format_result_details(details):
    if details is None:
        return None

    if not is_string(details):
        details = str(details)

    return details[0].upper() + details[1:]


def log_match_result(hint, matcher, result, quiet=False):
    """Add a check log to the report.

    If quiet is set to True, the check details won't appear in the check log.
    """
    description = "Expect %s %s" % (hint, matcher.description())

    return log_check(
        description, result.outcome, format_result_details(result.description) if not quiet else None
    )


def check_that(hint, actual, matcher, quiet=False):
    """Check that actual matches given matcher.

    A check log is added to the report.

    If quiet is set to True, the check details won't appear in the check log.
    """
    result = matcher.matches(actual)
    log_match_result(hint, matcher, result, quiet=quiet)
    return result.outcome

check_that_entry = _entry_operation(check_that)


def require_that(hint, actual, matcher, quiet=False):
    """Require that actual matches given matcher.

    A check log is added to the report. An AbortTest exception is raised if the check does not succeed.

    If quiet is set to True, the check details won't appear in the check log.
    """
    result = matcher.matches(actual)
    log_match_result(hint, matcher, result, quiet=quiet)
    if result.is_failure():
        raise AbortTest("previous requirement was not fulfilled")

require_that_entry = _entry_operation(require_that)


def assert_that(hint, actual, matcher, quiet=False):
    """Assert that actual matches given matcher.

    If assertion fail, a check log is added to the report and an AbortTest exception is raised.

    If quiet is set to True, the check details won't appear in the check log.
    """
    result = matcher.matches(actual)
    if result.is_failure():
        log_match_result(hint, matcher, result, quiet=quiet)
        raise AbortTest("assertion error")

assert_that_entry = _entry_operation(assert_that)



import time

from lemoncheesecake.utils import get_callable_args
from lemoncheesecake.exceptions import NoSuchEventType, MismatchingEventArguments
from lemoncheesecake.suite import Suite, Test
from lemoncheesecake.reporting.report import Report


class EventType:
    def __init__(self, event_type_name, event_arg_types):
        self._event_type_name = event_type_name
        self._event_arg_types = event_arg_types
        self._handlers = []

    def _has_extra_time_arg(self, args):
        return (len(args) == len(self._event_arg_types) + 1) and args[-1].endswith("time")

    def subscribe(self, handler):
        handler_args = get_callable_args(handler)

        handler_args_number = len(handler_args) - (1 if self._has_extra_time_arg(handler_args) else 0)
        if handler_args_number != len(self._event_arg_types):
            raise MismatchingEventArguments(
                "For event type '%s', expecting %d arguments, got %d in handler %s (not counting possible extra event_time argument)" % (
                    self._event_type_name, len(self._event_arg_types), handler_args_number, handler
            ))

        if self._has_extra_time_arg(handler_args):
            assert len(handler_args) == len(self._event_arg_types) + 1
        else:
            assert len(handler_args) == len(self._event_arg_types)

        self._handlers.append(handler)

    def unsubscribe(self, handler):
        self._handlers.remove(handler)

    def reset(self):
        self._handlers = []

    def _handle_event(self, handler, args, event_time):
        handler_expected_args = get_callable_args(handler)
        handler_args = list(args)
        if self._has_extra_time_arg(handler_expected_args):
            handler_args.append(event_time)
        handler(*handler_args)

    def fire(self, *args, **kwargs):
        event_time = kwargs.get("event_time") or time.time()

        # check that fire is called with the proper number of arguments
        if len(args) != len(self._event_arg_types):
            raise MismatchingEventArguments("For event type '%s', expecting %d arguments, got %d" % (
                self._event_type_name, len(self._event_arg_types), len(args)
            ))

        # check that fire is called with the appropriate argument types
        i = 0
        for handler_arg, event_arg in zip(args, self._event_arg_types):
            if (handler_arg is not None) and (not isinstance(handler_arg, event_arg)):
                raise MismatchingEventArguments("For event type '%s', expecting type '%s' for argument #%d, got '%s'" % (
                    self._event_type_name, event_arg, i+1, type(handler_arg)
                ))
            i += 1

        # call handlers
        for handler in self._handlers:
            self._handle_event(handler, args, event_time)


class EventManager:
    def __init__(self):
        self._event_types = {}

    def register_event_type(self, event_type_name, event_args):
        self._event_types[event_type_name] = EventType(event_type_name, event_args)
    
    def register_event_types(self, event_args, *event_type_names):
        for event_type_name in event_type_names:
            self.register_event_type(event_type_name, event_args)

    def reset(self, event_type_name=None):
        if event_type_name is None:
            for et in self._event_types.values():
                et.reset()
        else:
            self._call_event_type(event_type_name, lambda et: et.reset())

    def _call_event_type(self, event_type_name, callback):
        try:
            event_type = self._event_types[event_type_name]
        except KeyError:
            raise NoSuchEventType(event_type_name)
        callback(event_type)

    def subscribe_to_event_type(self, event_type_name, handler):
        self._call_event_type(event_type_name, lambda et: et.subscribe(handler))

    def subscribe_to_event_types(self, event_types):
        for event_type_name, handler in event_types.items():
            self.subscribe_to_event_type(event_type_name, handler)

    def unsubscribe_from_event_type(self, event_type_name, handler):
        self._call_event_type(event_type_name, lambda et: et.unsubscribe(handler))

    def fire(self, event_type_name, *args, **kwargs):
        self._call_event_type(event_type_name, lambda et: et.fire(*args, **kwargs))

    def add_listener(self, listener):
        for event_type_name in self._event_types.keys():
            try:
                sym = getattr(listener, event_type_name)
            except AttributeError:
                continue
            if not callable(sym):
                continue
            self.subscribe_to_event_type(event_type_name, sym)


eventmgr = EventManager()
register_event_type = eventmgr.register_event_type
register_event_types = eventmgr.register_event_types
subscribe_to_event_type = eventmgr.subscribe_to_event_type
subscribe_to_event_types = eventmgr.subscribe_to_event_types
unsubscribe_from_event_type = eventmgr.unsubscribe_from_event_type
add_listener = eventmgr.add_listener
reset = eventmgr.reset
fire = eventmgr.fire


###
# Global events
###

register_event_types(
    [Report],

    "on_tests_beginning", "on_tests_ending",
)

register_event_types(
    [],

    # test session setup & teardown beginning events
    "on_test_session_setup_beginning", "on_test_session_teardown_beginning"
)

register_event_types(
    [bool],

    # test session setup & teardown ending events
    "on_test_session_setup_ending", "on_test_session_teardown_ending",
)


###
# Suite events
###

register_event_types(
    [Suite],

    "on_suite_beginning", "on_suite_ending",
    "on_suite_setup_beginning",
    "on_suite_teardown_beginning"
)

register_event_types(
    [Suite, bool],

    "on_suite_setup_ending", "on_suite_teardown_ending"
)


###
# Test events
###

register_event_types(
    [Test],

    "on_test_beginning", "on_disabled_test",
    "on_test_setup_beginning", "on_test_teardown_beginning",
)

register_event_types(
    [Test, bool],

    "on_test_setup_ending", "on_test_teardown_ending"
)

register_event_type("on_skipped_test", [Test, str])
register_event_type("on_test_ending", [Test, str])


###
# Transverse test execution events
###

try:
    basestring
except NameError:
    # when using Python 3, just map basestring to str
    basestring = str

register_event_type("on_step", [basestring])
register_event_type("on_log", [str, basestring])
register_event_type("on_check", [basestring, bool, basestring])
register_event_type("on_log_attachment", [basestring, basestring, basestring])
register_event_type("on_log_url", [basestring, basestring])


'''
Created on Jan 24, 2016

@author: nicolas
'''

import time

from lemoncheesecake.consts import LOG_LEVEL_ERROR
from lemoncheesecake.reporting.report import *
from lemoncheesecake import events

__all__ = "ReportWriter", "initialize_report_writer"


class ReportWriter:
    def __init__(self, report):
        self.report = report
        # pointers to report data parts
        self.current_suite_data = None
        self.current_test_data = None
        self.current_step_data_list = None
        self.current_step_data = None
        # pointers to running test/suite
        self.current_test = None
        self.current_suite = None
        # for test / suite hook / before/after all tests outcome
        self.has_pending_failure = False
        # global status
        self._is_success = True

    def _start_hook(self, ts):
        self.has_pending_failure = False
        hook_data = HookData()
        hook_data.start_time = ts
        return hook_data

    def _end_hook(self, hook_data, ts):
        if hook_data:
            hook_data.end_time = ts
            hook_data.outcome = not self.has_pending_failure

    def end_current_step(self, ts):
        if self.current_step_data:
            self.current_step_data.end_time = ts
            self.current_step_data = None

        # remove previous step from report data if it was empty
        if self.current_step_data_list and len(self.current_step_data_list[-1].entries) == 0:
            del self.current_step_data_list[-1]

    def on_tests_beginning(self, report, start_time):
        self.report.start_time = start_time

    def on_tests_ending(self, report, end_time):
        self.report.end_time = end_time
        self.report.report_generation_time = self.report.end_time

    def on_test_session_setup_beginning(self, time):
        self.report.test_session_setup = self._start_hook(time)
        self.current_step_data_list = self.report.test_session_setup.steps

    def on_test_session_setup_ending(self, outcome, time):
        if self.report.test_session_setup.is_empty():
            self.report.test_session_setup = None
        else:
            self._end_hook(self.report.test_session_setup, time)
            self.end_current_step(time)

    def on_test_session_teardown_beginning(self, time):
        self.report.test_session_teardown = self._start_hook(time)
        self.current_step_data_list = self.report.test_session_teardown.steps

    def on_test_session_teardown_ending(self, outcome, time):
        if self.report.test_session_teardown.is_empty():
            self.report.test_session_teardown = None
        else:
            self._end_hook(self.report.test_session_teardown, time)
            self.end_current_step(time)

    def on_suite_beginning(self, suite):
        self.current_suite = suite
        suite_data = SuiteData(suite.name, suite.description)
        suite_data.tags.extend(suite.tags)
        suite_data.properties.update(suite.properties)
        suite_data.links.extend(suite.links)
        if self.current_suite_data:
            self.current_suite_data.add_suite(suite_data)
        else:
            self.report.add_suite(suite_data)
        self.current_suite_data = suite_data

    def on_suite_setup_beginning(self, suite, time):
        self.current_suite_data.suite_setup = self._start_hook(time)
        self.current_step_data_list = self.current_suite_data.suite_setup.steps

    def on_suite_setup_ending(self, suite, outcome, time):
        if self.current_suite_data.suite_setup.is_empty():
            self.current_suite_data.suite_setup = None
        else:
            self._end_hook(self.current_suite_data.suite_setup, time)
            self.end_current_step(time)

    def on_suite_teardown_beginning(self, suite, time):
        self.current_suite_data.suite_teardown = self._start_hook(time)
        self.current_step_data_list = self.current_suite_data.suite_teardown.steps

    def on_suite_teardown_ending(self, suite, outcome, time):
        if self.current_suite_data.suite_teardown.is_empty():
            self.current_suite_data.suite_teardown = None
        else:
            self.end_current_step(time)
            self._end_hook(self.current_suite_data.suite_teardown, time)

    def on_suite_ending(self, suite):
        self.current_suite_data = self.current_suite_data.parent_suite
        self.current_suite = self.current_suite.parent_suite

    def on_test_beginning(self, test, start_time):
        self.has_pending_failure = False
        self.current_test = test
        self.current_test_data = TestData(test.name, test.description)
        self.current_test_data.tags.extend(test.tags)
        self.current_test_data.properties.update(test.properties)
        self.current_test_data.links.extend(test.links)
        self.current_test_data.start_time = start_time
        self.current_suite_data.add_test(self.current_test_data)
        self.current_step_data_list = self.current_test_data.steps

    def on_test_ending(self, test, status, end_time):
        self.current_test_data.status = status
        self.current_test_data.end_time = end_time
        self.end_current_step(end_time)

        self.current_test = None
        self.current_test_data = None
        self.current_step_data_list = None

    def _bypass_test(self, test, status, status_details, time):
        test_data = TestData(test.name, test.description)
        test_data.tags.extend(test.tags)
        test_data.properties.update(test.properties)
        test_data.links.extend(test.links)
        test_data.end_time = test_data.start_time = time
        test_data.status = status
        test_data.status_details = status_details
        self.current_suite_data.add_test(test_data)

    def on_skipped_test(self, test, reason, time):
        self._is_success = False
        self._bypass_test(test, "skipped", reason, time)

    def on_disabled_test(self, test, time):
        self._bypass_test(test, "disabled", "", time)

    def on_step(self, description, step_time):
        self.end_current_step(step_time)

        self.current_step_data = StepData(description)
        self.current_step_data.start_time = step_time

        self.current_step_data_list.append(self.current_step_data)

    def on_log(self, level, content, log_time):
        if level == LOG_LEVEL_ERROR:
            self._is_success = False
            self.has_pending_failure = True
        self.current_step_data.entries.append(LogData(level, content, log_time))

    def on_check(self, description, outcome, details):
        self.current_step_data.entries.append(CheckData(description, outcome, details))

        if outcome is False:
            self._is_success = False
            self.has_pending_failure = True

    def on_log_attachment(self, path, filename, description):
        self.current_step_data.entries.append(AttachmentData(description, path))

    def on_log_url(self, url, description):
        self.current_step_data.entries.append(UrlData(description, url))

    def is_successful(self):
        return self._is_success


def initialize_report_writer(report):
    writer = ReportWriter(report)
    events.add_listener(writer)
    return writer



from .run import RunCommand
from .bootstrap import BootstrapCommand
from .show import ShowCommand
from .fixtures import FixturesCommand
from .stats import StatsCommand
from .report import ReportCommand
from .version import VersionCommand


def get_commands():
    return [
        RunCommand(), BootstrapCommand(),
        ShowCommand(), FixturesCommand(), StatsCommand(),
        ReportCommand(),
        VersionCommand()
    ]



'''
Created on Sep 8, 2016

@author: nicolas
'''

import fnmatch
from functools import reduce

from lemoncheesecake.exceptions import UserError

__all__ = ("Filter", "ReportFilter", "filter_suites", "add_filter_args_to_cli_parser", "make_filter_from_cli_args")

NEGATIVE_FILTER_CHARS = "-^~"


def match_values(values, patterns):
    if not patterns:
        return True

    values = [value or "" for value in values] # convert None to ""

    for pattern in patterns:
        if pattern[0] in NEGATIVE_FILTER_CHARS:
            if not fnmatch.filter(values, pattern[1:]):
                return True
        else:
            if fnmatch.filter(values, pattern):
                return True
    return False


def match_keyvalues(keyvalues, patterns):
    if not patterns:
        return True

    for key, value in patterns:
        if key in keyvalues:
            if value[0] in NEGATIVE_FILTER_CHARS:
                if not fnmatch.fnmatch(keyvalues[key], value[1:]):
                    return True
            else:
                if fnmatch.fnmatch(keyvalues[key], value):
                    return True
    return False


def match_values_lists(lsts, patterns):
    return match_values(
        reduce(lambda x, y: list(x) + list(y), lsts, []), # make a flat list
        patterns
    )


class Filter:
    def __init__(self):
        self.paths = []
        self.descriptions = []
        self.tags = []
        self.properties = []
        self.links = []
        self.enabled = False
        self.disabled = False

    def is_empty(self):
        return not any([
            self.paths, self.descriptions, self.tags, self.properties, self.links, self.enabled, self.disabled
        ])

    def is_test_disabled(self, test):
        return test.is_disabled()

    def match_test(self, test, suite):
        funcs = [
            lambda: self.is_test_disabled(test) if self.disabled else True,
            lambda: not self.is_test_disabled(test) if self.enabled else True,
            lambda: match_values(test.get_inherited_paths(), self.paths),
            lambda: all(match_values(test.get_inherited_descriptions(), descs) for descs in self.descriptions),
            lambda: all(match_values(test.get_inherited_tags(), tags) for tags in self.tags),
            lambda: all(match_keyvalues(test.get_inherited_properties(), props) for props in self.properties),
            lambda: all(match_values_lists(test.get_inherited_links(), links) for links in self.links)
        ]
        return all(func() for func in funcs)


class ReportFilter(Filter):
    def __init__(self):
        Filter.__init__(self)
        self.statuses = []

    def is_empty(self):
        if not Filter.is_empty(self):
            return False

        return len(self.statuses) == 0

    def is_test_disabled(self, test):
        return test.status == "disabled"

    def match_test(self, test, suite):
        if not Filter.match_test(self, test, suite):
            return False

        if len(self.statuses) == 0:
            return True

        return test.status in self.statuses


def filter_suite(suite, filtr):
    filtered_suite = suite.pull_node()

    for test in suite.get_tests():
        if filtr.match_test(test, suite):
            filtered_suite.add_test(test.pull_node())

    for filtered_sub_suite in filter_suites(suite.get_suites(), filtr):
        filtered_suite.add_suite(filtered_sub_suite)

    return filtered_suite


def filter_suites(suites, filtr):
    filtered_suites = []

    for suite in suites:
        filtered_suite = filter_suite(suite, filtr)
        if not filtered_suite.is_empty():
            filtered_suites.append(filtered_suite)

    return filtered_suites


def add_filter_args_to_cli_parser(cli_parser):
    def property_value(value):
        splitted = value.split(":")
        if len(splitted) != 2:
            raise ValueError()
        return splitted

    group = cli_parser.add_argument_group("Filtering")
    group.add_argument("path", nargs="*", default=[], help="Filter on test/suite path (wildcard character '*' can be used)")
    group.add_argument("--desc", nargs="+", action="append", default=[], help="Filter on descriptions")
    group.add_argument("--tag", "-a", nargs="+", action="append", default=[], help="Filter on tags")
    group.add_argument("--property", "-m", nargs="+", type=property_value, action="append", default=[], help="Filter on properties")
    group.add_argument("--link", "-l", nargs="+", action="append", default=[], help="Filter on links (names and URLs)")
    group.add_argument("--disabled", action="store_true", help="Filter on disabled tests")
    group.add_argument("--enabled", action="store_true", help="Filter on enabled (non-disabled) tests")

    return group


def add_report_filter_args_to_cli_parser(cli_parser):
    group = add_filter_args_to_cli_parser(cli_parser)
    group.add_argument("--passed", action="store_true", help="Filter on passed tests")
    group.add_argument("--failed", action="store_true", help="Filter on failed tests")
    group.add_argument("--skipped", action="store_true", help="Filter on skipped tests")


def _set_filter_from_cli_args(fltr, cli_args):
    if cli_args.disabled and cli_args.enabled:
        raise UserError("--disabled and --enabled arguments are mutually exclusive")

    fltr.paths = cli_args.path
    fltr.descriptions = cli_args.desc
    fltr.tags = cli_args.tag
    fltr.properties = cli_args.property
    fltr.links = cli_args.link
    fltr.disabled = cli_args.disabled
    fltr.enabled = cli_args.enabled

    return fltr


def make_filter_from_cli_args(cli_args):
    fltr = Filter()
    _set_filter_from_cli_args(fltr, cli_args)
    return fltr


def make_report_filter_from_cli_args(cli_args):
    fltr = ReportFilter()
    _set_filter_from_cli_args(fltr, cli_args)

    if cli_args.passed:
        fltr.statuses.append("passed")
    if cli_args.failed:
        fltr.statuses.append("failed")
    if cli_args.skipped:
        fltr.statuses.append("skipped")

    return fltr



'''
Created on Jan 24, 2016

@author: nicolas
'''

import os.path
import time
from contextlib import contextmanager
import shutil

from lemoncheesecake.exceptions import LemonCheesecakeInternalError
from lemoncheesecake.consts import ATTACHEMENT_DIR, \
    LOG_LEVEL_DEBUG, LOG_LEVEL_ERROR, LOG_LEVEL_INFO, LOG_LEVEL_WARN
from lemoncheesecake.reporting import *
from lemoncheesecake import events
from lemoncheesecake.exceptions import ProgrammingError

__all__ = "log_debug", "log_info", "log_warn", "log_warning", "log_error", "log_url", "log_check", \
    "set_step", "prepare_attachment", "save_attachment_file", "save_attachment_content", \
    "add_report_info", "get_fixture"


_runtime = None  # singleton


def initialize_runtime(report_dir, report, fixture_registry):
    global _runtime
    _runtime = _Runtime(report_dir, report, fixture_registry)
    events.add_listener(_runtime)


def get_runtime():
    if not _runtime:
        raise LemonCheesecakeInternalError("Runtime is not initialized")
    return _runtime


class _Runtime:
    def __init__(self, report_dir, report, fixture_registry):
        self.report_dir = report_dir
        self.report = report
        self.fixture_registry = fixture_registry
        self.attachments_dir = os.path.join(self.report_dir, ATTACHEMENT_DIR)
        self.attachment_count = 0
        self.step_lock = False
        self.pending_step = None
        self.pending_step_time = None

    def set_step(self, description, force_lock=False):
        if self.step_lock and not force_lock:
            return

        self.pending_step = description
        self.pending_step_time = time.time()

    def _flush_pending_step(self):
        if self.pending_step:
            events.fire("on_step", self.pending_step, event_time=self.pending_step_time)
            self.pending_step = None
            self.pending_step_time = None

    def log(self, level, content):
        self._flush_pending_step()
        events.fire("on_log", level, content)

    def log_check(self, description, outcome, details):
        self._flush_pending_step()
        events.fire("on_check", description, outcome, details)

    def log_url(self, url, description):
        self._flush_pending_step()
        events.fire("on_log_url", url, description)

    @contextmanager
    def prepare_attachment(self, filename, description):
        attachment_filename = "%04d_%s" % (self.attachment_count + 1, filename)
        self.attachment_count += 1
        if not os.path.exists(self.attachments_dir):
            os.mkdir(self.attachments_dir)

        yield os.path.join(self.attachments_dir, attachment_filename)

        self._flush_pending_step()
        events.fire("on_log_attachment", "%s/%s" % (ATTACHEMENT_DIR, attachment_filename), filename, description)

    def save_attachment_file(self, filename, description):
        with self.prepare_attachment(os.path.basename(filename), description) as report_attachment_path:
            shutil.copy(filename, report_attachment_path)

    def save_attachment_content(self, content, filename, description, binary_mode):
        with self.prepare_attachment(filename, description) as report_attachment_path:
            with open(report_attachment_path, "wb") as fh:
                fh.write(content if binary_mode else content.encode("utf-8"))

    def get_fixture(self, name):
        try:
            scope = self.fixture_registry.get_fixture_scope(name)
        except KeyError:
            raise ProgrammingError("Fixture '%s' does not exist" % name)

        if scope != "session_prerun":
            raise ProgrammingError("Fixture '%s' has scope '%s' while only fixtures with scope 'session_prerun' can be retrieved" % (
                name, scope
            ))

        try:
            return self.fixture_registry.get_fixture_result(name)
        except AssertionError as excp:
            raise ProgrammingError(str(excp))


def log_debug(content):
    """
    Log a debug level message.
    """
    get_runtime().log(LOG_LEVEL_DEBUG, content)


def log_info(content):
    """
    Log a info level message.
    """
    get_runtime().log(LOG_LEVEL_INFO, content)


def log_warning(content):
    """
    Log a warning level message.
    """
    get_runtime().log(LOG_LEVEL_WARN, content)

log_warn = log_warning


def log_error(content):
    """
    Log an error level message.
    """
    get_runtime().log(LOG_LEVEL_ERROR, content)


def log_check(description, outcome, details=None):
    get_runtime().log_check(description, outcome, details)


def set_step(description, force_lock=False):
    """
    Set a new step.
    """
    get_runtime().set_step(description, force_lock)


def prepare_attachment(filename, description=None):
    """
    Prepare a attachment using a pseudo filename and an optional description.
    The function returns the real filename on disk that will be used by the caller
    to write the attachment content.
    """
    return get_runtime().prepare_attachment(filename, description or filename)


def save_attachment_file(filename, description=None):
    """
    Save an attachment using an existing file (identified by filename) and an optional
    description. The given file will be copied.
    """
    get_runtime().save_attachment_file(filename, description or filename)


def save_attachment_content(content, filename, description=None, binary_mode=False):
    """
    Save a given content as attachment using pseudo filename and optional description.
    """
    get_runtime().save_attachment_content(content, filename, description or filename, binary_mode)


def log_url(url, description=None):
    """
    Log an URL.
    """
    get_runtime().log_url(url, description or url)


def get_fixture(name):
    """
    Return the corresponding fixture value. Only fixtures whose scope is session_prerun can be retrieved.
    """
    return get_runtime().get_fixture(name)


def add_report_info(name, value):
    report = get_runtime().report
    report.add_info(name, value)



'''
Created on Jun 16, 2017

@author: nicolas
'''

import copy

from lemoncheesecake.utils import get_distincts_in_list
from lemoncheesecake.exceptions import CannotFindTreeNode


class BaseTreeNode:
    def __init__(self, name, description):
        self.parent_suite = None
        self.name = name
        self.description = description
        self.tags = [ ]
        self.properties = {}
        self.links = [ ]

    def get_path(self):
        path = [ self ]
        parent_suite = self.parent_suite
        while parent_suite != None:
            path.insert(0, parent_suite)
            parent_suite = parent_suite.parent_suite
        return path
    
    def get_path_as_str(self, sep="."):
        return sep.join([s.name for s in self.get_path()])

    def get_depth(self):
        return len(self.get_path()) - 1

    def get_inherited_paths(self):
        return list(map(lambda node: node.get_path_as_str(), self.get_path()))

    def get_inherited_descriptions(self):
        return list(map(lambda node: node.description, self.get_path()))

    def get_inherited_tags(self):
        tags = []
        for node in self.get_path():
            tags.extend(node.tags)
        return get_distincts_in_list(tags)

    def get_inherited_properties(self):
        properties = {}
        for node in self.get_path():
            properties.update(node.properties)
        return properties

    def get_inherited_links(self):
        links = []
        for node in self.get_path():
            links.extend(node.links)
        return get_distincts_in_list(links)

    def pull_node(self):
        node = copy.copy(self)
        node.parent_suite = None
        return node

    def __str__(self):
        return self.get_path_as_str()


class BaseTest(BaseTreeNode):
    pass


class BaseSuite(BaseTreeNode):
    def __init__(self, name, description):
        BaseTreeNode.__init__(self, name, description)
        self._tests = []
        self._suites = []

    def add_test(self, test):
        test.parent_suite = self
        self._tests.append(test)

    def get_tests(self):
        return self._tests

    def add_suite(self, suite):
        suite.parent_suite = self
        self._suites.append(suite)

    def get_suites(self, include_empty_suites=False):
        if include_empty_suites:
            return self._suites
        else:
            return list(filter(lambda suite: not suite.is_empty(), self._suites))

    def is_empty(self):
        if len(self.get_tests()) != 0:
            return False

        for sub_suite in self.get_suites():
            if not sub_suite.is_empty():
                return False

        return True

    def pull_node(self):
        node = BaseTreeNode.pull_node(self)
        node._tests = []
        node._suites = []
        return node


def walk_suites(suites, suite_func=None, test_func=None):
    def do_walk(suite):
        if suite_func:
            suite_func(suite)
        if test_func:
            for test in suite.get_tests():
                test_func(test, suite)
        for sub_suite in suite.get_suites():
            do_walk(sub_suite)
    for suite in suites:
        do_walk(suite)


def walk_tests(suites, func):
    walk_suites(suites, test_func=func)


def get_flattened_suites(suites):
    flattened_suites = []
    walk_suites(suites, lambda suite: flattened_suites.append(suite))
    return flattened_suites


def get_suite_by_name(suites, suite_name):
    try:
        return next(s for s in suites if s.name == suite_name)
    except StopIteration:
        raise CannotFindTreeNode("Cannot find suite named '%s'" % suite_name)


def find_suite(suites, path, sep="."):
    lookup_suites = suites
    lookup_suite = None
    for lookup_suite_name in path.split(sep):
        lookup_suite = get_suite_by_name(lookup_suites, lookup_suite_name)
        lookup_suites = lookup_suite.get_suites(include_empty_suites=True)
    if lookup_suite is None:
        raise CannotFindTreeNode("Cannot find suite named '%s'" % path)

    return lookup_suite


def get_test_by_name(suite, test_name):
    try:
        return next(t for t in suite.get_tests() if t.name == test_name)
    except StopIteration:
        raise CannotFindTreeNode("Cannot find test named '%s'" % test_name)


def find_test(suites, path, sep="."):
    suite_name = sep.join(path.split(sep)[:-1])
    lookup_suite = find_suite(suites, suite_name, sep=sep)
    return get_test_by_name(lookup_suite, path.split(sep)[-1])


'''
Created on Jan 1, 2017

@author: nicolas
'''

import os
import platform
import time

def archive_dirname_datetime(ts, archives_dir):
    return time.strftime("report-%Y%m%d-%H%M%S", time.localtime(ts))

# TODO: create two different functions
def report_dir_with_archiving(top_dir, archive_dirname_callback):
    archives_dir = os.path.join(top_dir, "reports")

    if platform.system() == "Windows":
        report_dir = os.path.join(top_dir, "report")
        if os.path.exists(report_dir):
            if not os.path.exists(archives_dir):
                os.mkdir(archives_dir)
            os.rename(
                report_dir, os.path.join(archives_dir, archive_dirname_callback(os.path.getctime(report_dir), archives_dir))
            )
        os.mkdir(report_dir)

    else:
        if not os.path.exists(archives_dir):
            os.mkdir(archives_dir)

        report_dirname = archive_dirname_callback(time.time(), archives_dir)

        report_dir = os.path.join(archives_dir, report_dirname)
        os.mkdir(report_dir)

        symlink_path = os.path.join(os.path.dirname(report_dir), "..", "report")
        if os.path.lexists(symlink_path):
            os.unlink(symlink_path)
        os.symlink(report_dir, symlink_path)

    return report_dir



'''
Created on Sep 8, 2016

@author: nicolas
'''

import inspect
import copy

from lemoncheesecake.suite.loader import get_test_methods_from_class
from lemoncheesecake.suite.core import InjectedFixture
from lemoncheesecake.exceptions import ProgrammingError

__all__ = "add_test_in_suite", "add_tests_in_suite", "get_metadata", \
    "suite", "test", "tags", "prop", "link", "disabled", "conditional", "hidden", "inject_fixture"


class Metadata:
    _next_rank = 1

    def __init__(self):
        self.is_test = False
        self.is_suite = False
        self.name = None
        self.description = None
        self.properties = {}
        self.tags = []
        self.links = []
        self.rank = 0
        self.disabled = False
        self.condition = None


def get_metadata_next_rank():
    rank = Metadata._next_rank
    Metadata._next_rank += 1
    return rank


def add_test_in_suite(test, suite, before_test=None, after_test=None):
    # pre-checks
    if before_test and after_test:
        raise ProgrammingError("before_test and after_test are mutually exclusive")

    if hasattr(suite, test.name):
        raise ProgrammingError("Object %s has already an attribute named '%s'" % (suite, test.name))

    # build test func metadata
    md = test.callback._lccmetadata = Metadata()
    md.is_test = True
    md.name = test.name
    md.description = test.description
    md.tags.extend(test.tags)
    md.properties.update(test.properties)
    md.links.extend(test.links)
    md.disabled = test.disabled

    # set test func rank
    if before_test or after_test:
        ref_test_name = before_test if before_test else after_test
        test_methods = get_test_methods_from_class(suite)
        try:
            ref_test_obj = next(t for t in test_methods if t._lccmetadata.name == ref_test_name)
        except StopIteration:
            raise ProgrammingError("There is no base test named '%s' in class %s" % (ref_test_name, suite))

        if before_test:
            md.rank = ref_test_obj._lccmetadata.rank
            for test_method in filter(lambda t: t._lccmetadata.rank >= md.rank , test_methods):
                test_method._lccmetadata.rank += 1
        else: # after_test
            md.rank = ref_test_obj._lccmetadata.rank
            for test_method in filter(lambda t: t._lccmetadata.rank <= md.rank , test_methods):
                test_method._lccmetadata.rank -= 1
    else:
        md.rank = get_metadata_next_rank()

    # set test func and suite test method
    setattr(suite, test.name, test.callback.__get__(suite))


def add_tests_in_suite(tests, suite, before_test=None, after_test=None):
    if before_test and after_test:
        raise ProgrammingError("before_test and after_test are mutually exclusive")

    if after_test:
        previous_test = after_test
        for test in tests:
            add_test_in_suite(test, suite, after_test=previous_test)
            previous_test = test.name
    else:
        for test in tests:
            add_test_in_suite(test, suite, before_test=before_test)

_objects_with_metadata = []
def get_metadata(obj):
    global _objects_with_metadata

    if hasattr(obj, "_lccmetadata"):
        if obj not in _objects_with_metadata: # metadata comes from the superclass
            obj._lccmetadata = copy.deepcopy(obj._lccmetadata)
        return obj._lccmetadata
    else:
        obj._lccmetadata = Metadata()
        _objects_with_metadata.append(obj)
        return obj._lccmetadata


def suite(description, name=None, rank=None):
    """Decorator, mark a class as a suite class"""
    def wrapper(klass):
        if not inspect.isclass(klass):
            raise ProgrammingError("%s is not a class (suite decorator can only be used on a class)" % klass)
        md = get_metadata(klass)
        md.is_suite = True
        md.rank = rank if rank != None else get_metadata_next_rank()
        md.name = name or klass.__name__
        md.description = description
        return klass
    return wrapper


def test(description, name=None):
    """Decorator, make a method as a test method"""
    def wrapper(func):
        if not inspect.isfunction(func):
            raise ProgrammingError("%s is not a function (test decorator can only be used on a function)" % func)
        md = get_metadata(func)
        md.is_test = True
        md.rank = get_metadata_next_rank()
        md.name = name or func.__name__
        md.description = description
        return func
    return wrapper


def tags(*tag_names):
    """Decorator, add tags to a test or a suite"""
    def wrapper(obj):
        md = get_metadata(obj)
        md.tags.extend(tag_names)
        return obj
    return wrapper


def prop(key, value):
    """Decorator, add a property (key/value) to a test or a suite"""
    def wrapper(obj):
        md = get_metadata(obj)
        md.properties[key] = value
        return obj
    return wrapper


def link(url, name=None):
    """Decorator, set a link (with an optional friendly name) to a test or a suite"""
    def wrapper(obj):
        md = get_metadata(obj)
        md.links.append((url, name))
        return obj
    return wrapper


def disabled():
    """Decorator, disable a test or a suite"""
    def wrapper(obj):
        md = get_metadata(obj)
        md.disabled = True
        return obj
    return wrapper


def conditional(condition):
    """Decorator, the test or suite will only appear if the given condition callback return a true value"""
    def wrapper(obj):
        md = get_metadata(obj)
        md.condition = condition
        return obj
    return wrapper


def hidden():
    return conditional(lambda _: False)


def inject_fixture(fixture_name=None):
    return InjectedFixture(fixture_name)


from __future__ import print_function

from lemoncheesecake.cli.command import Command
from lemoncheesecake.reporting import load_report
from lemoncheesecake.project import find_project_file, load_project_from_file
from lemoncheesecake.reporting.backends.console import display_report_suites
from lemoncheesecake.filter import add_report_filter_args_to_cli_parser, make_report_filter_from_cli_args, filter_suites
from lemoncheesecake.exceptions import ProjectError


class ReportCommand(Command):
    def get_name(self):
        return "report"

    def get_description(self):
        return "Display a report"

    def add_cli_args(self, cli_parser):
        group = cli_parser.add_argument_group("Show report")
        group.add_argument("report_path", help="Report file name or directory")
        add_report_filter_args_to_cli_parser(cli_parser)

    def run_cmd(self, cli_args):
        project_filename = find_project_file()
        if project_filename is not None:
            try:
                project = load_project_from_file(project_filename)
            except ProjectError:
                # project is optional when reading a report
                report_backends = None
            else:
                report_backends = project.get_all_reporting_backends()
        else:
            report_backends = None

        report = load_report(cli_args.report_path, report_backends)
        suites = filter_suites(report.get_suites(), make_report_filter_from_cli_args(cli_args))

        display_report_suites(suites)

        return 0



'''
Created on Mar 12, 2017

@author: nicolas
'''

from __future__ import print_function

import lemoncheesecake
from lemoncheesecake.filter import make_filter_from_cli_args, filter_suites
from lemoncheesecake.exceptions import UserError

LEMONCHEESECAKE_VERSION = "lemoncheesecake version %s" % lemoncheesecake.__version__


def filter_suites_from_cli_args(suites, cli_args):
    filtr = make_filter_from_cli_args(cli_args)
    if filtr.is_empty():
        return suites

    suites = filter_suites(suites, filtr)
    if len(suites) == 0:
        raise UserError("The filter does not match any test")

    return suites


def get_suites_from_project(project, cli_args):
    suites = project.get_suites()
    if all(suite.is_empty() for suite in suites):
        raise UserError("No test is defined in your lemoncheesecake project.")

    return filter_suites_from_cli_args(suites, cli_args)
